/**---------------------------------------------------------------------------

  Name         :  lcd_EG9018C.c
  Description  :  STM32 driver for EPSON EG9018C passive matrix LCD
  Author       :  Aleksander Borysiuk

  History      :  2013.07.16 - ver.1.0

------------------------------------------------------------------------------*/

/**---------------------------------------------------------------------------
  Hardware info:

  EPSON EG9018C is a passive matrix graphic monochrome LCD with VGA resolution
  e.a. 640x480 pixels. The display uses 30 pin ZIFF connector designed for 1mm
  FFC cable.
                    ----------------------------
                   |   EG9018C ZIFF connector   |
                    ----------------------------
  1  - Data 0 Upper                       16 - GND
  2  - CCFL brightness control            17 - CP2 / Data input clock
  3  - Data 1 Upper                       18 - GND
  4  - GND                                19 - CP1 / Input data latch
  5  - Data 2 Upper                       20 - S / Scan start-up
  6  - VEE control (LCD contrast)         21 - GND
  7  - Data 3 Upper                       22 - +5V power supply for VEE inverter
  8  - GND                                23 - +5V power supply for VEE inverter
  9  - Data 0 Lower                       24 - +5V power supply
  10 - CCFL inverter ON                   25 - +5V power supply
  11 - Data 1 Lower                       26 - GND
  12 - GND                                27 - pin 3 for J2203 connector
  13 - Data 2 Lower                       28 - pin 1 for J2203 connector
  14 - VEE inverter ON                    29 - pin 2 for J2203 connector
  15 - Data 3 Lower                       30 - pin 4 for J2203 connector

  EPSON EG9018C display module has no built-in LCD controller. All its control signals
  must be provided by external CPU.

  STM32 microcontroller services EG9018C LCD in hardware way only without any software
  burden put on CPU. LCD control signals are generated by STM32 timers while data transfer
  to LCD is performed by DMA. Details about used STM32 peripherals and their operation
  are given in header comments of EG9018C_Init() function.

------------------------------------------------------------------------------*/

#include "stm32f10x_conf.h"
#include "lcd_EG9018C.h"
#include "fonts.h"


/* Private define ------------------------------------------------------------*/
/* Private constants for EPSON EG9018C LCD module ----------------------------*/
#define EG9018C_DATA  GPIOD     // LCD data bus: PD7 -> DU3, PD6 -> DU2, PD5 -> DU1, PD4 -> DU0,
                                //               PD3 -> DL3, PD2 -> DL2, PD1 -> DL1, PD0 -> DL0

#define EG9018C_CP2_PERIOD   28 // LCD data clock CP2 period (fCP2 = 2,571MHz i.e. 66,5 frames/s)

#define EG9018C_CP1_PERIOD  160 // LCD data latch CP1 period (TCP1 = 160 pulses of CP1 clock)
#define EG9018C_CP1_LEN       1 // LCD data latch CP1 length (tCP1 = 1 pulse of CP1 clock i.e. 0,389us)

#define EG9018C_S_PERIOD  38400 // LCD start-up S period (TS = 240*160 = 38400 pulses of CP1 clock)
#define EG9018C_S_LEN       166 // LCD start-up S length (tS = 166 pulses of CP1 clock i.e. 64,55us)


/* Private variables ---------------------------------------------------------*/
uint8_t EG9018C_frame_buffer [EG9018C_YRES/2][ EG9018C_XRES/4]; // LCD frame buffer

/* Private function prototypes -----------------------------------------------*/

/* Private functions ---------------------------------------------------------*/

/* Public functions ----------------------------------------------------------*/

/**-----------------------------------------------------------------------------
  Name         :  EG9018C_Init
  Description  :  Function configures STM32 peripherals for generating EPSON
                  EG9018C LCD control signals
  Argument(s)  :  none
  Return value :  none
--------------------------------------------------------------------------------
  Details:

  EPSON EG9018C LCD is serviced by STM32 hardware only. No software burden is put
  on CPU. The LCD driver uses the following STM32 peripherals:
  - TIM4                 -> LCD data clock generator
  - TIM4_CH1 on pin PD12 -> LCD CP2 signal (i.e. data clock output)
  - TIM5                 -> LCD data latch generator
  - TIM5_CH1 on pin PA0  -> LCD CP1 signal (i.e. data latch output)
  - TIM1                 -> LCD start-up timer
  - TIM1_CH1 on pin PA8  -> LCD S signal (i.e. start-up output)
  - PD7..PD0             -> LCD data DU4..DU0,DL4..DL0
  - DMA1_CH1             -> LCD data transfer engine
  - TIM5_CH2 on pin PA1  -> LCD CONTRAST adjustment
  - TIM4_CH2 on pin PD13 -> LCD BACKLIGHT on/off
  - PA2                  -> LCD CONTRAST VOLTAGE GENERATOR enable

  STM32 peripherals work as follows:
  - EG9018C data clock CP2 is generated on pin PD12 by channel 1 of timer TIM4.
    TIM4 works in PWM1 mode and it is clocked by APB1 clock.
  - EG9018C data latch CP1 is generated on pin PA0 by channel 1 of timer TIM5.
    TIM5 works in PWM1 mode and it is clocked by timer TIM4 update event.
  - EG9018C start-up S is generated on pin PA8 by channel 1 of timer TIM1.
    TIM1 works in PWM1 mode and it is clocked by timer TIM4 update event.
    Start-up delay is provided by TIM1 OCx output dead time generator.
  - EG9018C data is sent to pins PD7..PD0 by DMA1 controller via channel 1.
    DMA transfer is requested by compare event in channel 1 of timer TIM4.
  - EG9018C contrast is regulated by PWM signal that is generated on pin PA1
    by channel 2 of timer TIM5.
  - EG9018C backlight is controlled by PWM signal that is generated on pin PD13
    by channel 2 of timer TIM4.
  - EG9018C contrast voltage generator, if software control is enabled, is activated
    by high level on pin PA2
------------------------------------------------------------------------------*/
void EG9018C_Init(void)
{
  GPIO_InitTypeDef GPIO_InitStructure;
  TIM_TimeBaseInitTypeDef TIM_TimeBaseStructure;
  TIM_OCInitTypeDef TIM_OCInitStructure;
  TIM_BDTRInitTypeDef TIM_BDTRInitStructure;
  DMA_InitTypeDef DMA_InitStructure;

  /// --------------------------------------------------------------------------
  ///                      Clock enable of used peripherals
  ///         EG9018C uses DMA1, TIM1, TIM4, TIM5, GPIOA, GPIOD and AFIO
  /// --------------------------------------------------------------------------
  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_DMA1, ENABLE);
  RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM4 | RCC_APB1Periph_TIM5, ENABLE);
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOA | RCC_APB2Periph_GPIOD | \
                         RCC_APB2Periph_AFIO | RCC_APB2Periph_TIM1, ENABLE);


  /// --------------------------------------------------------------------------
  ///                Configuration of LCD data lines
  ///       EG9018C data lines are connected to pins PD7..PD0
  /// --------------------------------------------------------------------------
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_7 | GPIO_Pin_6 | GPIO_Pin_5 | GPIO_Pin_4 |\
                                GPIO_Pin_3 | GPIO_Pin_2 | GPIO_Pin_1 | GPIO_Pin_0;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_Out_PP;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_Init(EG9018C_DATA, &GPIO_InitStructure);


  /// --------------------------------------------------------------------------
  ///           Configuration of LCD data clock CP2 and BACKLIGHT on/off
  /// EG9018C data clock CP2 is generated on pin PD12 by channel 1 of timer TIM4
  /// Backlight on/off sigal is generated on pin PD13 by channel 2 of timer TIM4
  /// --------------------------------------------------------------------------
  // PD12 configuration as TIM4_CH1 output, PD13 configuration as TIM4_CH2 output
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_12 | GPIO_Pin_13;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF_PP;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_Init(GPIOD, &GPIO_InitStructure);
  GPIO_PinRemapConfig(GPIO_Remap_TIM4, ENABLE);

  // TIM4 time base configuration (f_clk=72MHz)
  TIM_TimeBaseStructure.TIM_Period = EG9018C_CP2_PERIOD-1;
  TIM_TimeBaseStructure.TIM_Prescaler = 0;
  TIM_TimeBaseStructure.TIM_ClockDivision = 0;
  TIM_TimeBaseStructure.TIM_CounterMode = TIM_CounterMode_Up;
  TIM_TimeBaseInit(TIM4, &TIM_TimeBaseStructure);
  TIM_ARRPreloadConfig(TIM4, ENABLE);

  // TIM4 Channel1 configuration: PWM1 Mode / 50%
  TIM_OCInitStructure.TIM_OCMode = TIM_OCMode_PWM1;
  TIM_OCInitStructure.TIM_Pulse = EG9018C_CP2_PERIOD/2;
  TIM_OCInitStructure.TIM_OutputState = TIM_OutputState_Enable;
  TIM_OCInitStructure.TIM_OCPolarity = TIM_OCPolarity_High;
  TIM_OC1Init(TIM4, &TIM_OCInitStructure);
  TIM_OC1PreloadConfig(TIM4, TIM_OCPreload_Enable);

  // TIM4 Channel2 configuration: PWM1 Mode / 0% or 100%
  TIM_OCInitStructure.TIM_OCMode = TIM_OCMode_PWM1;
  TIM_OCInitStructure.TIM_Pulse = EG9018C_BKLT_OFF;
  TIM_OCInitStructure.TIM_OutputState = TIM_OutputState_Enable;
  TIM_OCInitStructure.TIM_OCPolarity = TIM_OCPolarity_High;
  TIM_OC2Init(TIM4, &TIM_OCInitStructure);
  TIM_OC2PreloadConfig(TIM4, TIM_OCPreload_Enable);

  // Clock source configuration: TIM4 is the master for timers TIM2 and TIM5
  TIM_SelectMasterSlaveMode(TIM4, TIM_MasterSlaveMode_Enable);
  TIM_SelectOutputTrigger(TIM4, TIM_TRGOSource_Update);


  /// --------------------------------------------------------------------------
  ///        Configuration of LCD data latch CP1 and CONTRAST adjustment
  /// EG9018C data latch CP1 is generated on pin PA0 by channel 1 of timer TIM5
  /// Contrast adjustment voltage is generated on pin PA1 by channel 2 of timer TIM5
  /// --------------------------------------------------------------------------
  // PA0 configuration as TIM5_CH1 output, PA1 configuration as TIM5_CH2 output
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_0 | GPIO_Pin_1;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF_PP;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_Init(GPIOA, &GPIO_InitStructure);

  // Clock source configuration: TIM5 is the slave of TIM4
  TIM_SelectSlaveMode(TIM5, TIM_SlaveMode_External1);
  TIM_SelectInputTrigger(TIM5, TIM_TS_ITR2);

  // TIM5 time base configuration (f_clk=fTIM4)
  TIM_TimeBaseStructure.TIM_Period = EG9018C_CP1_PERIOD-1;
  TIM_TimeBaseStructure.TIM_Prescaler = 0;
  TIM_TimeBaseStructure.TIM_ClockDivision = 0;
  TIM_TimeBaseStructure.TIM_CounterMode = TIM_CounterMode_Up;
  TIM_TimeBaseInit(TIM5, &TIM_TimeBaseStructure);
  TIM_ARRPreloadConfig(TIM5, ENABLE);

  // TIM5 Channel1 configuration: PWM1 Mode
  TIM_OCInitStructure.TIM_OCMode = TIM_OCMode_PWM1;
  TIM_OCInitStructure.TIM_Pulse = EG9018C_CP1_PERIOD-EG9018C_CP1_LEN;
  TIM_OCInitStructure.TIM_OutputState = TIM_OutputState_Enable;
  TIM_OCInitStructure.TIM_OCPolarity = TIM_OCPolarity_Low;
  TIM_OC1Init(TIM5, &TIM_OCInitStructure);
  TIM_OC1PreloadConfig(TIM5, TIM_OCPreload_Enable);

  // TIM5 Channel2 configuration: PWM1 Mode
  TIM_OCInitStructure.TIM_OCMode = TIM_OCMode_PWM1;
  TIM_OCInitStructure.TIM_Pulse = EG9018C_CNTR_ADJ;
  TIM_OCInitStructure.TIM_OutputState = TIM_OutputState_Enable;
  TIM_OCInitStructure.TIM_OCPolarity = TIM_OCPolarity_High;
  TIM_OC2Init(TIM5, &TIM_OCInitStructure);
  TIM_OC2PreloadConfig(TIM5, TIM_OCPreload_Enable);


  /// --------------------------------------------------------------------------
  ///                  Configuration of LCD start-up S
  /// EG9018C start-up S is generated on pin PA8 by channel 1 of timer TIM1
  /// Start-up pulse delay is provided by TIM1 OCx output dead time generator
  /// --------------------------------------------------------------------------
  // PA8 configuration as TIM1_CH1 output
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_8;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF_PP;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_Init(GPIOA, &GPIO_InitStructure);

  // Clock source configuration: TIM1 is the slave of TIM4
  TIM_SelectSlaveMode(TIM1, TIM_SlaveMode_External1);
  TIM_SelectInputTrigger(TIM1, TIM_TS_ITR3);

  // TIM1 time base configuration (f_clk=fTIM4)
  TIM_TimeBaseStructure.TIM_Period = EG9018C_S_PERIOD-1;
  TIM_TimeBaseStructure.TIM_Prescaler = 0;
  TIM_TimeBaseStructure.TIM_ClockDivision = 0;
  TIM_TimeBaseStructure.TIM_CounterMode = TIM_CounterMode_Up;
  TIM_TimeBaseStructure.TIM_RepetitionCounter = 0;
  TIM_TimeBaseInit(TIM1, &TIM_TimeBaseStructure);
  TIM_ARRPreloadConfig(TIM1, ENABLE);

  // TIM1 Channel1 configuration: PWM1 Mode
  TIM_OCInitStructure.TIM_OCMode = TIM_OCMode_PWM1;
  TIM_OCInitStructure.TIM_Pulse = EG9018C_S_LEN;
  TIM_OCInitStructure.TIM_OutputState = TIM_OutputState_Enable;
  TIM_OCInitStructure.TIM_OCPolarity = TIM_OCPolarity_High;
  TIM_OCInitStructure.TIM_OCIdleState = TIM_OCIdleState_Reset;
  TIM_OCInitStructure.TIM_OutputNState = TIM_OutputNState_Enable;
  TIM_OCInitStructure.TIM_OCNPolarity = TIM_OCNPolarity_High;
  TIM_OCInitStructure.TIM_OCNIdleState = TIM_OCNIdleState_Reset;
  TIM_OC1Init(TIM1, &TIM_OCInitStructure);
  TIM_OC1PreloadConfig(TIM1, TIM_OCPreload_Enable);

  // TIM1 Dead Time
  TIM_BDTRInitStructure.TIM_DeadTime = 0xFF;
  TIM_BDTRInitStructure.TIM_AutomaticOutput = TIM_AutomaticOutput_Disable;
  TIM_BDTRInitStructure.TIM_Break = TIM_Break_Disable;
  TIM_BDTRInitStructure.TIM_BreakPolarity = TIM_BreakPolarity_Low;
  TIM_BDTRInitStructure.TIM_LOCKLevel = TIM_LOCKLevel_OFF;
  TIM_BDTRInitStructure.TIM_OSSIState = TIM_OSSIState_Disable;
  TIM_BDTRInitStructure.TIM_OSSRState = TIM_OSSRState_Disable;
  TIM_BDTRConfig (TIM1,&TIM_BDTRInitStructure);

  // TIM1 Main Outputs Enable
  TIM_CtrlPWMOutputs(TIM1, ENABLE);


  /// --------------------------------------------------------------------------
  ///                  Configuration of LCD data transfer
  /// EG9018C data is transfered to pins PD7..PD0 by DMA1 controller via channel 1
  /// DMA transfer is requested by compare event in channel 1 of timer TIM4
  /// --------------------------------------------------------------------------
  // DMA1 Channel 1 configuration
  DMA_DeInit(DMA1_Channel1);

  DMA_InitStructure.DMA_PeripheralBaseAddr = (uint32_t)&(EG9018C_DATA->ODR);
  DMA_InitStructure.DMA_MemoryBaseAddr = (uint32_t)&EG9018C_frame_buffer[0][0];
  DMA_InitStructure.DMA_DIR = DMA_DIR_PeripheralDST;
  DMA_InitStructure.DMA_BufferSize = (EG9018C_XRES/4)*(EG9018C_YRES/2);
  DMA_InitStructure.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
  DMA_InitStructure.DMA_MemoryInc = DMA_MemoryInc_Enable;
  DMA_InitStructure.DMA_PeripheralDataSize = DMA_PeripheralDataSize_Word;
  DMA_InitStructure.DMA_MemoryDataSize = DMA_MemoryDataSize_Byte;
  DMA_InitStructure.DMA_Mode = DMA_Mode_Circular;
  DMA_InitStructure.DMA_Priority = DMA_Priority_VeryHigh;
  DMA_InitStructure.DMA_M2M = DMA_M2M_Disable;
  DMA_Init(DMA1_Channel1, &DMA_InitStructure);

  // Enable DMA1 Channel 1
  DMA_Cmd(DMA1_Channel1, ENABLE);

  // Set compare event in channel 1 of timer TIM4 to start DMA transfer
  TIM_DMACmd(TIM4, TIM_DMA_CC1, ENABLE);


  /// --------------------------------------------------------------------------
  ///                      CONTRAST VOLTAGE GENERATOR enable
  /// EG9018C contrast voltage generator is enabled by high level on pin PA2
  /// --------------------------------------------------------------------------
#if (EG9018C_CNTR_GEN_SWD == 1)
  // PA2 configuration as output
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_2;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_Out_PP;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_2MHz;
  GPIO_Init(GPIOA, &GPIO_InitStructure);

  // enable EG9018C contrast voltage generator
  GPIO_SetBits(GPIOA, GPIO_Pin_2);
#endif

  /// --------------------------------------------------------------------------
  ///                         Start servicing LCD
  /// --------------------------------------------------------------------------
  // Request the first DMA transfer
  TIM_GenerateEvent(TIM4, TIM_EventSource_CC1);

  // Enable TIM1 counter
  TIM_Cmd(TIM1, ENABLE);

  // Enable TIM5 counter
  TIM_Cmd(TIM5, ENABLE);

  // Enable TIM4  counter
  TIM_Cmd(TIM4, ENABLE);
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_Contrast
  Description  :  Function sets LCD contrast
  Argument(s)  :  contrast (0 - max. contrast to 160 - min. contrast)
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_Contrast(uint8_t contrast)
{
  TIM_SetCompare2(TIM5, contrast);
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_Backlight
  Description  :  Function controls LCD backlight
  Argument(s)  :  bkltState - backlight state: EG9018C_BKLT_OFF or EG9018C_BKLT_ON
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_Backlight(uint8_t bkltState)
{
  TIM_SetCompare2(TIM4, bkltState);
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_PixelOn
  Description  :  Function draws pixel
  Argument(s)  :  Xpos -> pixel x coordinate (0..639)
                  Ypos -> pixel y coordinate (0..479)
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_PixelOn(uint16_t Xpos, uint16_t Ypos)
{
  uint8_t colNo, rowNo, bitNo;                      // column number, row number and bit number in frame buffer

  if (Xpos < EG9018C_XRES && Ypos < EG9018C_YRES){  // if pixel coordinates within range

    // Calculate pixel position in LCD frame buffer
    colNo = Xpos/4;
    rowNo = (Ypos < EG9018C_YRES/2) ? Ypos : Ypos - EG9018C_YRES/2;
    bitNo = (Ypos < EG9018C_YRES/2) ? 7 - Xpos%4 : 3 - Xpos%4;

    // Draw pixel
    // (acces to pixel bit is performed via bit-banding region of SRAM)
    *(__IO uint32_t *)(SRAM_BB_BASE | (((uint32_t)&EG9018C_frame_buffer[rowNo][colNo] - SRAM_BASE) << 5) \
                       | ((bitNo) << 2)) = LCD_PIXEL_ON;
  }
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_PixelOff
  Description  :  Function clears pixel
  Argument(s)  :  Xpos -> pixel x coordinate (0..639)
                  Ypos -> pixel y coordinate (0..479)
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_PixelOff(uint16_t Xpos, uint16_t Ypos)
{
  uint8_t colNo, rowNo, bitNo;                      // column number, row number and bit number in frame buffer

  if (Xpos < EG9018C_XRES && Ypos < EG9018C_YRES){  // if pixel coordinates within range

    // Calculate pixel position in LCD frame buffer
    colNo = Xpos/4;
    rowNo = (Ypos < EG9018C_YRES/2) ? Ypos : Ypos - EG9018C_YRES/2;
    bitNo = (Ypos < EG9018C_YRES/2) ? 7 - Xpos%4 : 3 - Xpos%4;

    // Clear pixel
    // (acces to pixel bit is performed via bit-banding region of SRAM)
    *(__IO uint32_t *)(SRAM_BB_BASE | (((uint32_t)&EG9018C_frame_buffer[rowNo][colNo] - SRAM_BASE) << 5) \
                       | ((bitNo) << 2)) = LCD_PIXEL_OFF;
  }
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_Clear
  Description  :  Function clears LCD frame buffer
  Argument(s)  :  none
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_Clear(void)
{
  uint8_t i, j;                              // loop row and column counters

   for (i=0; i<EG9018C_YRES/2; i++){
    for (j=0; j<EG9018C_XRES/4; j++){
      EG9018C_frame_buffer[i][j] = (LCD_PIXEL_OFF << 7 | LCD_PIXEL_OFF << 6 | \
                                    LCD_PIXEL_OFF << 5 | LCD_PIXEL_OFF << 4 | \
                                    LCD_PIXEL_OFF << 3 | LCD_PIXEL_OFF << 2 |
                                    LCD_PIXEL_OFF << 1 | LCD_PIXEL_OFF);
    }
  }
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_DrawLine
  Description  :  Function draws line
  Argument(s)  :  Xpos -> line x start coordinate (0..639)
                  Ypos -> line y start coordinate (0..479)
                  length -> line length
                  direction -> line direction (LCD_HORIZONTAL or LCD_VERTICAL)
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_DrawLine(uint16_t Xpos, uint16_t Ypos, uint16_t length, uint8_t direction)
{
  uint16_t i;                                // pixel counter

  if (direction == LCD_HORIZONTAL){          // horizontal line
    for (i=0; i < length; i++){
      LCD_PixelOn(Xpos++, Ypos);             // draw pixel
    }
  }
  else {                                     // vertical line
    for (i=0; i < length; i++){
      LCD_PixelOn(Xpos, Ypos++);             // draw pixel
    }
  }
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_DrawRect
  Description  :  Function draws a rectangle
  Argument(s)  :  Xpos -> x start coordinate (0..639)
                  Ypos -> y start coordinate (0..479)
                  width  -> rectangle width
                  height -> rectangle height
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_DrawRect(uint16_t Xpos, uint16_t Ypos, uint16_t width, uint16_t height)
{
  LCD_DrawLine(Xpos, Ypos, width, LCD_HORIZONTAL);
  LCD_DrawLine(Xpos, (Ypos + height - 1), width, LCD_HORIZONTAL);

  LCD_DrawLine(Xpos, Ypos, height, LCD_VERTICAL);
  LCD_DrawLine((Xpos + width - 1), Ypos, height, LCD_VERTICAL);
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_DrawCircle
  Description  :  Function draws a circle using Midpoint algorithm
  Argument(s)  :  Xpos -> center x coordinate (0..639)
                  Ypos -> center y coordinate (0..479)
                  radius -> circle radius
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_DrawCircle(uint16_t Xcenter, uint16_t Ycenter, uint16_t radius)
{
  int32_t  x;                                // current x value
  int32_t  y;                                // current y value
  int32_t  p;                                // decision variable (p<0 - point inside the circle,
                                             // p>0 - point outside the circle)

  x = 0;                                     // circle start point x coordinate
  y = radius;                                // circle start point y coordinate
  p = (5 - (radius << 2)) >> 2;              // p = (5 - 4*radius) / 4

  while (x <= y){
    // x=0 any y=0 axis symetry
    LCD_PixelOn(Xcenter + x, Ycenter + y);
    LCD_PixelOn(Xcenter - x, Ycenter + y);
    LCD_PixelOn(Xcenter + x, Ycenter - y);
    LCD_PixelOn(Xcenter - x, Ycenter - y);

    // y=x and y=-x axis symetry
    LCD_PixelOn(Xcenter + y, Ycenter + x);
    LCD_PixelOn(Xcenter - y, Ycenter + x);
    LCD_PixelOn(Xcenter + y, Ycenter - x);
    LCD_PixelOn(Xcenter - y, Ycenter - x);

    if (p < 0){                              // if point inside the circle
      p += (x << 1) + 1;                     // p = p + 2*x + 1
    }
    else {                                   // if point outside the circle
      p += ((x - y) << 1) + 1;               // p = p + 2*(x - y) + 1
      y--;
    }
    x++;                                     // the next x coordinate
  }
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_DrawPicture
  Description  :  Function draws mono picture at given location.
                  The picture pixels are processed in rows, starting from msb in data
                  byte. If horizontal picture size is not multiple of 8, the remaining
                  bits of data byte are discarded prior to starting the new row.
  Argument(s)  :  Xpos  -> picture top left corner x coordinate (0..639)
                  Ypos  -> picture top left corner y coordinate (0..479)
                  Xsize -> picture horizontal size in pixels (640 max)
                  Ysize -> picture vertical size in pixels (480 max)
                  *p_pict -> pointer to picture data table
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_DrawPicture (uint16_t Xpos, uint16_t Ypos, uint16_t Xsize, uint16_t Ysize, uint8_t *p_pict)
{
  uint16_t  x, y;                            // current x and y values
  uint8_t   i;                               // bit counter
  uint8_t   pictData;                        // picture data


  x = 0;                                     // picture start coordinate x
  y = 0;                                     // picture start coordinate y

  while (y < Ysize){
    pictData = *(p_pict++);                  // get picture data

    for (i=0; i < 8; i++){                   // process picture data byte
      if (pictData & 0x80){                  // if pixel black
        LCD_PixelOn(Xpos + x, Ypos + y);     // draw pixel
      }
      else {                                 // if pixel white
        LCD_PixelOff(Xpos + x, Ypos + y);    // clear pixel
      }
      pictData <<= 1;                        // prepare the next pixel

      x++;                                   // the next x coordinate
      if (x >= Xsize){                       // if end of row
        x = 0;                               // start new row
        y++;                                 // the next row
        break;                               // new row starts with new data byte
      }
    }
  }
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_DrawGlyph
  Description  :  Function draws specified letter (called "a glyph" in "fonts.c"
                  library) at location x, y
  Argument(s)  :  Xpos        -> glyph top left corner x coordinate (0..639)
                  Ypos        -> glyph top left corner y coordinate (0..479)
                  width       -> glyph horizontal size in pixels
                  height      -> glyph vertical size in pixels
                  *p_glyph    -> pointer to glyph pixels
                  store_width -> glyph pattern store width
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_DrawGlyph(uint16_t Xpos, uint16_t Ypos, uint8_t width, uint8_t height,
                   uint8_t *p_glyph, uint8_t store_width)
{
  uint16_t x, y;                             // current x and y values
  uint8_t  i;                                // bit counter
  uint8_t  glyphData;                        // glyph data
  uint8_t  *p_glyphScan;                     // glyph scan pointer


  x = 0;                                     // glyph start coordinate x
  y = 0;                                     // glyph start coordinate y
  p_glyphScan = p_glyph;                     // glyph scan start

  while (y < height){
    glyphData = *(p_glyphScan++);            // get glyph data

    for (i=0; i < 8; i++){                   // process glyph data byte
      if (glyphData & 0x80){                 // if pixel black
        LCD_PixelOn(Xpos + x, Ypos + y);     // draw pixel
      }
      else {                                 // if pixel white
        LCD_PixelOff(Xpos + x, Ypos + y);    // clear pixel
      }
      glyphData <<= 1;                       // prepare the next pixel

      x++;                                   // the next x coordinate
      if (x >= width){                       // if end of row
        x = 0;                               // start new row
        y++;                                 // the next row
        p_glyph += store_width;              // move glyph pointer to the next row
        p_glyphScan = p_glyph;               // set glyph scan start to the begining of row
        break;                               // new row starts with new data byte
      }
    }
  }
}


/**-----------------------------------------------------------------------------
  Name         :  LCD_DrawText
  Description  :  Function displays text at given location using specified font.
                  Each character of the text is displayed via function call that
                  draws specified letter (called "a glyph" in "fonts.c" library).
                  If text contains character undefined in specified font set
                  (typically covering ASCI codes 0x20..0x7E), this character is
                  substituted with default letter.
  Argument(s)  :  Xpos  -> text top left corner x coordinate (0..639)
                  Ypos  -> text top left corner y coordinate (0..479)
                  font  -> font set (FONT_FIVE_DOT, FONT_SIX_DOT, FONT_SEVEN_DOT,
                                     FONT_NINE_DOT, FONT_TEN_DOT, FONT_FIFTEEN_DOT,
                                     FONT_EIGHTEEN_DOT, FONT_FONT_MT1
                  *p_text -> pointer to null terminated text string
  Return value :  none
------------------------------------------------------------------------------*/
void LCD_DrawText (uint16_t Xpos, uint16_t Ypos, uint8_t font, uint8_t *p_text)
{
  uint8_t letter;                            // letter code
  uint8_t width;                             // letter pattern width
  uint8_t height;                            // letter pattern height
  uint8_t store_width;                       // letter pattern store width
  uint8_t *p_glyph;                          // letter pattern pointer


  while (*p_text != 0x00){                   // repeat untill end of the text
    letter = *p_text;                        // get character

    // check if the letter pattern is included in font set
    if((letter < fonts[font].glyph_beg) || (letter > fonts[font].glyph_end)){
      letter = fonts[font].glyph_def;        // replace letter with default character
    }

    // calculate pointer to the letter pattern
    letter -= fonts[font].glyph_beg;         // remove font patterns table offset

    width = fonts[font].fixed_width;         // font pattern width
	if(width == 0){                          // if font pattern width is variable
	  width=fonts[font].width_table[letter]; // get font pattern width from width table
	}
	store_width = fonts[font].store_width;   // font pattern store width

	height = fonts[font].glyph_height;       // font pattern height

    p_glyph = (uint8_t*) fonts[font].glyph_table + ((uint16_t)letter * (uint16_t)store_width * (uint16_t)height);

    // draw the letter (glyph)
    LCD_DrawGlyph(Xpos, Ypos, width, height, p_glyph, store_width);

    Xpos += width;                           // move right for the next letter
    p_text++;                                // poin to the next letter in string
  }
}

/*----------------------------- End of file ----------------------------------*/
